#include <stdio.h>
#include <shmem.h>
#include "pgmio.h"

#define nx 360
#define ny 240
#define numpes 8
#define nxlocal (nx/numpes)
#define nylocal ny

double bigpic[nx][ny];
double oldpic[nxlocal+2][nylocal+2];
double newpic[nxlocal+2][nylocal+2];

int main(void)
{
  
  int ix, iy;
  int pe, mype, rtpe, ltpe;

  shmem_init();

  mype = shmem_my_pe();

  if (numpes != shmem_n_pes())
  {

    if (mype == 0)
    {
      printf("Error: compiled for %d PE(s), but running on %d\n", numpes, shmem_n_pes());
    }

    shmem_finalize();

    return -1;

  }

  if (mype == 0)
  {
    printf("Running with %d PE(s)\n",numpes);
  }

  printf("Hello from PE %d out of %d\n",mype,numpes);

  for (ix=0; ix < (nxlocal+2); ix++)
  {
    for (iy=0; iy < (nylocal+2); iy++)
    {
      oldpic[ix][iy] = 0.0;
      newpic[ix][iy] = 0.0;
    }  
  }

  /* Need to synchronise to ensure smallpic is initialised on all PEs
     before the master PE reads them. */

  shmem_barrier_all();

  /* Read the picture on the master and copy to every PE's oldpic */

  if (mype == 0)
  {

    pgmread("picture360x240.pgm", bigpic, nx, ny);

  /* Mapping done in strips with multiple put calls as oldpic has a halo.
     A single block put would overlap that halo. */

    for (pe=0; pe < numpes; pe++)
    {
      for(ix = pe*nxlocal; ix < (pe+1)*nxlocal; ix++)
      {
        shmem_double_put(&oldpic[ (ix % nxlocal)+1 ][1], &bigpic[ix][0], nylocal, pe);
      }
    }

  }

  /* Overwrite bigpic to ensure that we have to do the output correctly */
  for (ix=0; ix < nx; ix++)
  {
    for (iy=0; iy < ny; iy++)
    {
      bigpic[ix][iy] = -1.0;
    }
  }

  /* Need to synchronise to ensure the master has completed initialising
     oldpic on all PEs before any of them try to read the halos. */

  shmem_barrier_all();

  /* Compute indices of neighbouring PEs */
  rtpe = mype + 1;
  ltpe = mype - 1;

  /* Do halo swap, taking care at the picture boundaries */
  
  if (rtpe < numpes)
  {
    shmem_double_get(&oldpic[nxlocal+1][1], &oldpic[1][1], nylocal, rtpe);
  }

  if (ltpe >= 0)
  {
    shmem_double_get(&oldpic[0][1], &oldpic[nxlocal][1], nylocal, ltpe);
  }

  /* Now compute the edges and place the answer in newpic */
  
  for (ix=1; ix <= nxlocal; ix++)
  {
    for (iy=1; iy <= nylocal; iy++)
    {
      newpic[ix][iy] =   oldpic[ix-1][iy] + oldpic[ix+1][iy]
                       + oldpic[ix][iy-1] + oldpic[ix][iy+1]
                       - 4.0 * oldpic[ix][iy];
    }
  }

  /* Now write the output on the master PE.

     Need to synchronise to ensure calculation of newpic is complete on
     all PEs before the master reads the values */

  shmem_barrier_all();
  
  if (mype == 0)
  {

    /* Copy newpic back to bigpic */
    for (pe=0; pe < numpes; pe++)
    {
      for(ix = pe*nxlocal; ix < (pe+1)*nxlocal; ix++)
      {
        shmem_double_get(&bigpic[ix][0], &newpic[ (ix % nxlocal)+1 ][1], nylocal, pe);
      }
    }

    pgmwrite("output.pgm", bigpic, nx, ny);

  }

  shmem_finalize();

  return 0;
}

