program ex2a

  use pgmio

  implicit none

  include 'shmem.fh'

  integer, parameter :: nx = 360
  integer, parameter :: ny = 240

  integer, parameter :: numpes = 8

  integer, parameter :: nxlocal = nx
  integer, parameter :: nylocal = ny/numpes

  integer :: ix, iy
  integer :: pe, mype, uppe, dnpe

  double precision, save :: bigpic(nx, ny)

  double precision, save :: oldpic(0:nxlocal+1, 0:nylocal+1)
  double precision, save :: newpic(0:nxlocal+1, 0:nylocal+1)

  call shmem_init()

  mype = shmem_my_pe()

  if (numpes /= shmem_n_pes()) then

    if (mype == 0) then
      write(*,*) 'Error: compiled for ', numpes, &
                 ' PE(s), but running on ', shmem_n_pes()
    end if

    call shmem_finalize()

    stop

  end if

! Even in Fortran, SHMEM PEs index from 0

  if (mype == 0)  write(*,*) 'Running on ', numpes, ' PE(s)'

  write(*,*) 'Hello from PE ', mype, ' out of ', numpes

  oldpic(:,:) = 0.0
  newpic(:,:) = 0.0 

! Need to synchronise to ensure all PEs have finished initialising
! their local pictures before the master writes to them.

  call shmem_barrier_all()

! Read the picture on the master and copy to every PE's oldpic

  if (mype == 0) then

    call pgmread('picture360x240.pgm', bigpic)

    do pe = 0, numpes-1

! Commented to avoid issues from array parameter being copied into a temp holder.
! Potential for non-blocking shmem operation to fail due to temporary array being cleared before 'send' is completed.
!      call shmem_double_put(oldpic(1:nxlocal,1:nylocal), bigpic(1,pe*nylocal+1), nxlocal*nylocal, pe)
!
! Mapping done in strips with multiple put calls as oldpic has a halo. A single block put would overlap that halo.

! Mod to map bigpic y index to smaller local oldpic
 
      do iy = pe*nylocal+1, (pe+1)*nylocal
        call shmem_double_put(oldpic(1,MOD(iy-1,nylocal)+1), bigpic(1,iy), nxlocal, pe)
      end do

    end do

  end if

! Overwrite bigpic to ensure that we have to do the output correctly

  bigpic(:,:) = -1.0

! Need to synchronise to ensure the master has completed initialising
! oldpic on all PEs before any of them try to read the halos.

  call shmem_barrier_all()

! Compute indices of neighbouring PEs

  uppe = mype + 1
  dnpe = mype - 1

! Do halo swap, taking care at the upper and lower picture boundaries

  if (uppe < numpes) then

    call shmem_double_get(oldpic(1, nylocal+1), oldpic(1, 1), nxlocal, uppe)

  end if

  if (dnpe >= 0) then

    call shmem_double_get(oldpic(1, 0), oldpic(1, nylocal), nxlocal, dnpe)

  end if

! Now compute the edges and place the answer in newpic

  do ix = 1, nxlocal
    do iy = 1, nylocal

      newpic(ix, iy) =   oldpic(ix-1, iy) + oldpic(ix+1, iy) &
                       + oldpic(ix, iy-1) + oldpic(ix, iy+1) &
                       - 4.0 * oldpic(ix,iy)

    end do
  end do

! Now write the output on the master PE.
!
! Need to synchronise to ensure calculation of newpic is complete on
! all PEs before the master reads the values.

  call shmem_barrier_all()

  if (mype == 0) then  

! Copy newpic back to bigpic

    do pe = 0, numpes-1

      do iy = pe*nylocal+1, (pe+1)*nylocal
        call shmem_double_get(bigpic(1,iy), newpic(1,MOD(iy-1,nylocal)+1), nxlocal, pe)
      end do

    end do

    call pgmwrite('output.pgm', bigpic)

  end if

  call shmem_finalize()

end program ex2a
