program ex2b

  use pgmio

  implicit none

  include 'shmem.fh'

  integer, parameter :: nx = 360
  integer, parameter :: ny = 240

  integer, parameter :: numpes = 8

  integer, parameter :: nxlocal = nx
  integer, parameter :: nylocal = ny/numpes

  integer, parameter :: maxloop = 100

  integer :: ix, iy, iloop
  integer :: pe, mype, uppe, dnpe

  double precision, save :: bigpic(nx, ny)

  double precision, save :: oldpic(0:nxlocal+1, 0:nylocal+1)
  double precision, save :: newpic(0:nxlocal+1, 0:nylocal+1)
  double precision, save ::   edge(0:nxlocal+1, 0:nylocal+1)

  call shmem_init()

  mype = shmem_my_pe()

  if (numpes /= shmem_n_pes()) then

    if (mype == 0) then
      write(*,*) 'Error: compiled for ', numpes, & 
                 ' PE(s), but running on ', shmem_n_pes()
    end if

    call shmem_finalize()

    stop

  end if

  if (mype == 0)  then

    write(*,*) 'Running on ', numpes, ' PE(s)'
    write(*,*)
    write(*,*) 'Performing ', maxloop, ' iterations'
    write(*,*)

  end if

  write(*,*) 'Hello from PE ', mype, ' out of ', numpes

  oldpic(:,:) = 0.0
  newpic(:,:) = 0.0
    edge(:,:) = 0.0

! Need to synchronise to ensure all PEs have finished initialising
! their local pictures before the master writes to them.

  call shmem_barrier_all()

! Read the picture on the master and copy to every PE's edge

  if (mype == 0) then

    call pgmread('edge360x240.pgm', bigpic)

    do pe = 0, numpes-1

! Mapping done in strips with multiple put calls as edge has a halo. A single block put would overlap that halo.
      do iy = pe*nylocal+1, (pe+1)*nylocal
        call shmem_double_put(edge(1,MOD(iy-1,nylocal)+1), bigpic(1,iy), nxlocal, pe)
      end do

    end do

  end if

! Overwrite bigpic to ensure that we have to do the output correctly

  bigpic(:,:) = -1.0

! Need to synchronise to ensure the master has completed initialising
! edge on all PEs before any of them try to read it.

  call shmem_barrier_all()

  oldpic(:,:) = edge(:,:)

! Compute indices of neighbouring PEs

  uppe = mype + 1
  dnpe = mype - 1

! Need to synchronise to ensure oldpic has been initialised on all PEs
! before any PEs try to access its halos on their neighbours

  call shmem_barrier_all()

  do iloop = 1, maxloop

! Do halo swap, taking care at the upper and lower picture boundaries

    if (uppe < numpes) then

      call shmem_double_get(oldpic(1, nylocal+1), oldpic(1, 1), nxlocal, uppe)

    end if

    if (dnpe >= 0) then

      call shmem_double_get(oldpic(1, 0), oldpic(1, nylocal), nxlocal, dnpe)

    end if

! Now update the local values of newpic

    do ix = 1, nxlocal
      do iy = 1, nylocal


        newpic(ix, iy) = 0.25*(   oldpic(ix-1, iy) + oldpic(ix+1, iy) &
                                + oldpic(ix, iy-1) + oldpic(ix, iy+1) &
                                - edge(ix,iy)                           )

      end do
    end do

! Need to synchronise to ensure that all PEs have finished reading the
! oldpic halo values on this PE before overwriting it with newpic

    call shmem_barrier_all() 

! Prepare for next iteration by setting oldpic = newpic

    oldpic(1:nxlocal,1:nylocal) = newpic(1:nxlocal,1:nylocal)

! Need to synchronise to ensure that all PEs have finished updating
! their oldpic arrays before this PE reads any halo data from them at
! the start of the next iteration

    call shmem_barrier_all()

  end do

! Now write the output on the master PE.
! No synchronisation required as due to sync all at the end of each loop

  if (mype == 0) then  

! Copy newpic back to bigpic

    do pe = 0, numpes-1

      do iy = pe*nylocal+1, (pe+1)*nylocal
        call shmem_double_get(bigpic(1,iy), newpic(1,MOD(iy-1,nylocal)+1), nxlocal, pe)
      end do

    end do

    call pgmwrite('output.pgm', bigpic)

  end if

! Missing this gives "Error detected during page fault processing.  Process terminated via bus error."
  call shmem_finalize()

end program ex2b
